package sun.util.locale.provider;

import java.text.StringCharacterIterator;
import java.security.PrivilegedActionException;
import java.security.AccessController;
import java.security.PrivilegedExceptionAction;
import java.io.BufferedInputStream;
import java.util.MissingResourceException;
import java.io.IOException;
import java.text.CharacterIterator;
import sun.text.SupplementaryCharacterData;
import sun.text.CompactByteArray;
import java.text.BreakIterator;

class RuleBasedBreakIterator extends BreakIterator
{
    protected static final byte IGNORE = -1;
    private static final short START_STATE = 1;
    private static final short STOP_STATE = 0;
    static final byte[] LABEL;
    static final int LABEL_LENGTH;
    static final byte supportedVersion = 1;
    private static final int HEADER_LENGTH = 36;
    private static final int BMP_INDICES_LENGTH = 512;
    private CompactByteArray charCategoryTable;
    private SupplementaryCharacterData supplementaryCharCategoryTable;
    private short[] stateTable;
    private short[] backwardsStateTable;
    private boolean[] endStates;
    private boolean[] lookaheadStates;
    private byte[] additionalData;
    private int numCategories;
    private CharacterIterator text;
    private long checksum;
    private int cachedLastKnownBreak;
    
    RuleBasedBreakIterator(String s) throws IOException, MissingResourceException {
        this.charCategoryTable = null;
        this.supplementaryCharCategoryTable = null;
        this.stateTable = null;
        this.backwardsStateTable = null;
        this.endStates = null;
        this.lookaheadStates = null;
        this.additionalData = null;
        this.text = null;
        this.cachedLastKnownBreak = -1;
        this.readTables(s);
    }
    
    protected final void readTables(String s) throws IOException, MissingResourceException {
        byte[] file = this.readFile(s);
        int int1 = getInt(file, 0);
        int int2 = getInt(file, 4);
        int int3 = getInt(file, 8);
        int int4 = getInt(file, 12);
        int int5 = getInt(file, 16);
        int int6 = getInt(file, 20);
        int int7 = getInt(file, 24);
        this.checksum = getLong(file, 28);
        this.stateTable = new short[int1];
        int n = 36;
        for (int i = 0; i < int1; ++i, n += 2) {
            this.stateTable[i] = getShort(file, n);
        }
        this.backwardsStateTable = new short[int2];
        for (int j = 0; j < int2; ++j, n += 2) {
            this.backwardsStateTable[j] = getShort(file, n);
        }
        this.endStates = new boolean[int3];
        for (int k = 0; k < int3; ++k, ++n) {
            this.endStates[k] = (file[n] == 1);
        }
        this.lookaheadStates = new boolean[int4];
        for (int l = 0; l < int4; ++l, ++n) {
            this.lookaheadStates[l] = (file[n] == 1);
        }
        short[] indexArray = new short[512];
        for (int n2 = 0; n2 < 512; ++n2, n += 2) {
            indexArray[n2] = getShort(file, n);
        }
        byte[] newValues = new byte[int5];
        System.arraycopy(file, n, newValues, 0, int5);
        int n3 = n + int5;
        this.charCategoryTable = new CompactByteArray(indexArray, newValues);
        int[] table = new int[int6];
        for (int n4 = 0; n4 < int6; ++n4, n3 += 4) {
            table[n4] = getInt(file, n3);
        }
        this.supplementaryCharCategoryTable = new SupplementaryCharacterData(table);
        if (int7 > 0) {
            System.arraycopy(file, n3, this.additionalData = new byte[int7], 0, int7);
        }
        this.numCategories = this.stateTable.length / this.endStates.length;
    }
    
    protected byte[] readFile(final String s) throws IOException, MissingResourceException {
        BufferedInputStream bufferedInputStream;
        try {
            bufferedInputStream = (BufferedInputStream)AccessController.doPrivileged(new PrivilegedExceptionAction<BufferedInputStream>() {
                @Override
                public BufferedInputStream run() throws Exception {
                    return new BufferedInputStream(this.getClass().getResourceAsStream("/sun/text/resources/" + s));
                }
            });
        }
        catch (PrivilegedActionException cause) {
            throw new InternalError(cause.toString(), cause);
        }
        int n = 0;
        int n2 = RuleBasedBreakIterator.LABEL_LENGTH + 5;
        byte[] b = new byte[n2];
        if (bufferedInputStream.read(b) != n2) {
            throw new MissingResourceException("Wrong header length", s, "");
        }
        int i = 0;
        while (i < RuleBasedBreakIterator.LABEL_LENGTH) {
            if (b[n] != RuleBasedBreakIterator.LABEL[n]) {
                throw new MissingResourceException("Wrong magic number", s, "");
            }
            ++i;
            ++n;
        }
        if (b[n] != 1) {
            throw new MissingResourceException("Unsupported version(" + (int)b[n] + ")", s, "");
        }
        int int1 = getInt(b, ++n);
        byte[] b2 = new byte[int1];
        if (bufferedInputStream.read(b2) != int1) {
            throw new MissingResourceException("Wrong data length", s, "");
        }
        bufferedInputStream.close();
        return b2;
    }
    
    byte[] getAdditionalData() {
        return this.additionalData;
    }
    
    void setAdditionalData(byte[] additionalData) {
        this.additionalData = additionalData;
    }
    
    @Override
    public Object clone() {
        RuleBasedBreakIterator ruleBasedBreakIterator = (RuleBasedBreakIterator)super.clone();
        if (this.text != null) {
            ruleBasedBreakIterator.text = (CharacterIterator)this.text.clone();
            return ruleBasedBreakIterator;
        }
        return ruleBasedBreakIterator;
    }
    
    @Override
    public boolean equals(Object o) {
        try {
            if (o == null) {
                return false;
            }
            RuleBasedBreakIterator ruleBasedBreakIterator = (RuleBasedBreakIterator)o;
            if (this.checksum != ruleBasedBreakIterator.checksum) {
                return false;
            }
            if (this.text == null) {
                return ruleBasedBreakIterator.text == null;
            }
            return this.text.equals(ruleBasedBreakIterator.text);
        }
        catch (ClassCastException ex) {
            return false;
        }
    }
    
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append('[');
        sb.append("checksum=0x");
        sb.append(Long.toHexString(this.checksum));
        sb.append(']');
        return sb.toString();
    }
    
    @Override
    public int hashCode() {
        return (int)this.checksum;
    }
    
    @Override
    public int first() {
        CharacterIterator text = this.getText();
        text.first();
        return text.getIndex();
    }
    
    @Override
    public int last() {
        CharacterIterator text = this.getText();
        text.setIndex(text.getEndIndex());
        return text.getIndex();
    }
    
    @Override
    public int next(int i) {
        int n = this.current();
        while (i > 0) {
            n = this.handleNext();
            --i;
        }
        while (i < 0) {
            n = this.previous();
            ++i;
        }
        return n;
    }
    
    @Override
    public int next() {
        return this.handleNext();
    }
    
    @Override
    public int previous() {
        CharacterIterator text = this.getText();
        if (this.current() == text.getBeginIndex()) {
            return -1;
        }
        int current = this.current();
        int cachedLastKnownBreak = this.cachedLastKnownBreak;
        if (cachedLastKnownBreak >= current || cachedLastKnownBreak <= -1) {
            this.getPrevious();
            cachedLastKnownBreak = this.handlePrevious();
        }
        else {
            text.setIndex(cachedLastKnownBreak);
        }
        for (; cachedLastKnownBreak != -1 && cachedLastKnownBreak < current; , cachedLastKnownBreak = this.handleNext()) {}
        text.setIndex(cachedLastKnownBreak);
        return this.cachedLastKnownBreak = cachedLastKnownBreak;
    }
    
    private int getPrevious() {
        int previous = this.text.previous();
        if (Character.isLowSurrogate((char)previous) && this.text.getIndex() > this.text.getBeginIndex()) {
            int previous2 = this.text.previous();
            if (Character.isHighSurrogate((char)previous2)) {
                return Character.toCodePoint((char)previous2, (char)previous);
            }
            this.text.next();
        }
        return previous;
    }
    
    int getCurrent() {
        int current = this.text.current();
        if (Character.isHighSurrogate((char)current) && this.text.getIndex() < this.text.getEndIndex()) {
            int next = this.text.next();
            this.text.previous();
            if (Character.isLowSurrogate((char)next)) {
                return Character.toCodePoint((char)current, (char)next);
            }
        }
        return current;
    }
    
    private int getCurrentCodePointCount() {
        if (Character.isHighSurrogate(this.text.current()) && this.text.getIndex() < this.text.getEndIndex()) {
            int next = this.text.next();
            this.text.previous();
            if (Character.isLowSurrogate((char)next)) {
                return 2;
            }
        }
        return 1;
    }
    
    int getNext() {
        int index = this.text.getIndex();
        int endIndex = this.text.getEndIndex();
        int index2;
        if (index == endIndex || (index2 = index + this.getCurrentCodePointCount()) >= endIndex) {
            return 65535;
        }
        this.text.setIndex(index2);
        return this.getCurrent();
    }
    
    private int getNextIndex() {
        int n = this.text.getIndex() + this.getCurrentCodePointCount();
        int endIndex = this.text.getEndIndex();
        if (n > endIndex) {
            return endIndex;
        }
        return n;
    }
    
    protected static final void checkOffset(int n, CharacterIterator characterIterator) {
        if (n < characterIterator.getBeginIndex() || n > characterIterator.getEndIndex()) {
            throw new IllegalArgumentException("offset out of bounds");
        }
        return;
    }
    
    @Override
    public int following(int index) {
        CharacterIterator text = this.getText();
        checkOffset(index, text);
        text.setIndex(index);
        if (index == text.getBeginIndex()) {
            return this.cachedLastKnownBreak = this.handleNext();
        }
        int n = this.cachedLastKnownBreak;
        if (n >= index || n <= -1) {
            n = this.handlePrevious();
        }
        else {
            text.setIndex(n);
        }
        while (n != -1 && n <= index) {
            n = this.handleNext();
        }
        return this.cachedLastKnownBreak = n;
    }
    
    @Override
    public int preceding(int index) {
        CharacterIterator text = this.getText();
        checkOffset(index, text);
        text.setIndex(index);
        return this.previous();
    }
    
    @Override
    public boolean isBoundary(int n) {
        CharacterIterator text = this.getText();
        checkOffset(n, text);
        return n == text.getBeginIndex() || this.following(n - 1) == n;
    }
    
    @Override
    public int current() {
        return this.getText().getIndex();
    }
    
    @Override
    public CharacterIterator getText() {
        if (this.text == null) {
            this.text = new StringCharacterIterator("");
        }
        return this.text;
    }
    
    @Override
    public void setText(CharacterIterator text) {
        int endIndex = text.getEndIndex();
        int n;
        try {
            text.setIndex(endIndex);
            n = (text.getIndex() == endIndex);
        }
        catch (IllegalArgumentException ex) {
            n = false;
        }
        if (!(n == 0)) {
            this.text = text;
        }
        else {
            this.text = new SafeCharIterator(text);
        }
        this.text.first();
        this.cachedLastKnownBreak = -1;
    }
    
    protected int handleNext() {
        CharacterIterator text = this.getText();
        if (text.getIndex() == text.getEndIndex()) {
            return -1;
        }
        int index = this.getNextIndex();
        int nextIndex = 0;
        int lookupState;
        int n;
        for (lookupState = 1, n = this.getCurrent(); n != 65535 && lookupState != 0; n = this.getNext()) {
            int lookupCategory = this.lookupCategory(n);
            if (lookupCategory != -1) {
                lookupState = this.lookupState(lookupState, lookupCategory);
            }
            if (this.lookaheadStates[lookupState]) {
                if (this.endStates[lookupState]) {
                    index = nextIndex;
                }
                else {
                    nextIndex = this.getNextIndex();
                }
            }
            else if (this.endStates[lookupState]) {
                index = this.getNextIndex();
            }
        }
        if (n == 65535 && nextIndex == text.getEndIndex()) {
            index = nextIndex;
        }
        text.setIndex(index);
        return index;
    }
    
    protected int handlePrevious() {
        CharacterIterator text = this.getText();
        int lookupBackwardState = 1;
        int lookupCategory = 0;
        int n = 0;
        int n2;
        for (n2 = this.getCurrent(); n2 != 65535 && lookupBackwardState != 0; n2 = this.getPrevious()) {
            n = lookupCategory;
            lookupCategory = this.lookupCategory(n2);
            if (lookupCategory != -1) {
                lookupBackwardState = this.lookupBackwardState(lookupBackwardState, lookupCategory);
            }
        }
        if (n2 != 65535) {
            if (n != -1) {
                this.getNext();
                this.getNext();
            }
            else {
                this.getNext();
            }
        }
        return text.getIndex();
    }
    
    protected int lookupCategory(int index) {
        if (index < 65536) {
            return this.charCategoryTable.elementAt((char)index);
        }
        return this.supplementaryCharCategoryTable.getValue(index);
    }
    
    protected int lookupState(int n, int n2) {
        return this.stateTable[n * this.numCategories + n2];
    }
    
    protected int lookupBackwardState(int n, int n2) {
        return this.backwardsStateTable[n * this.numCategories + n2];
    }
    
    static long getLong(byte[] array, int n) {
        long n2 = (long)(array[n] & 0xFF);
        for (int i = 1; i < 8; ++i) {
            n2 = (n2 << 8 | (long)(array[n + i] & 0xFF));
        }
        return n2;
    }
    
    static int getInt(byte[] array, int n) {
        int n2 = array[n] & 0xFF;
        for (int i = 1; i < 4; ++i) {
            n2 = (n2 << 8 | (array[n + i] & 0xFF));
        }
        return n2;
    }
    
    static short getShort(byte[] array, int n) {
        return (short)((short)(array[n] & 0xFF) << 8 | (array[n + 1] & 0xFF));
    }
    
    static {
        LABEL = new byte[] { 66, 73, 100, 97, 116, 97, 0 };
        LABEL_LENGTH = RuleBasedBreakIterator.LABEL.length;
    }
    
    private static final class SafeCharIterator implements CharacterIterator, Cloneable
    {
        private CharacterIterator base;
        private int rangeStart;
        private int rangeLimit;
        private int currentIndex;
        
        SafeCharIterator(CharacterIterator base) {
            this.base = base;
            this.rangeStart = base.getBeginIndex();
            this.rangeLimit = base.getEndIndex();
            this.currentIndex = base.getIndex();
        }
        
        @Override
        public char first() {
            return this.setIndex(this.rangeStart);
        }
        
        @Override
        public char last() {
            return this.setIndex(this.rangeLimit - 1);
        }
        
        @Override
        public char current() {
            if (this.currentIndex < this.rangeStart || this.currentIndex >= this.rangeLimit) {
                return '\uffff';
            }
            return this.base.setIndex(this.currentIndex);
        }
        
        @Override
        public char next() {
            ++this.currentIndex;
            if (this.currentIndex >= this.rangeLimit) {
                this.currentIndex = this.rangeLimit;
                return '\uffff';
            }
            return this.base.setIndex(this.currentIndex);
        }
        
        @Override
        public char previous() {
            --this.currentIndex;
            if (this.currentIndex < this.rangeStart) {
                this.currentIndex = this.rangeStart;
                return '\uffff';
            }
            return this.base.setIndex(this.currentIndex);
        }
        
        @Override
        public char setIndex(int currentIndex) {
            if (currentIndex < this.rangeStart || currentIndex > this.rangeLimit) {
                throw new IllegalArgumentException("Invalid position");
            }
            this.currentIndex = currentIndex;
            return this.current();
        }
        
        @Override
        public int getBeginIndex() {
            return this.rangeStart;
        }
        
        @Override
        public int getEndIndex() {
            return this.rangeLimit;
        }
        
        @Override
        public int getIndex() {
            return this.currentIndex;
        }
        
        @Override
        public Object clone() {
            RuleBasedBreakIterator.SafeCharIterator safeCharIterator;
            try {
                safeCharIterator = (RuleBasedBreakIterator.SafeCharIterator)super.clone();
            }
            catch (CloneNotSupportedException obj) {
                throw new Error("Clone not supported: " + obj);
            }
            safeCharIterator.base = (CharacterIterator)this.base.clone();
            return safeCharIterator;
        }
    }
}
